uniform float   Time;
uniform float   ObjScale;
uniform float   Spin;

const vec3      testPosition    = vec3( 0.0, 0.0, 40.0 );
const vec3      testNormal      = vec3( 0.0, 1.0, 0.0 );
const float     SpecularContribution = 0.3;
const float     DiffuseContribution = 1.0 - SpecularContribution;

// parameters for the wave function
const float     xdiv = 10.0;
const float     zdiv = 8.0;
const float     magx = 1.6;
const float     magz = 1.0;

mat4      identity = mat4 (   1.0, 0.0, 0.0, 0.0,
                              0.0, 1.0, 0.0, 0.0,
                              0.0, 0.0, 1.0, 0.0,
                              0.0, 0.0, 0.0, 1.0);

varying float   LightIntensity;
varying vec3    MCposition;
varying vec3    normal;

vec3            tn;
vec3            tp;
mat4            mvmat; // to be used instead of modelview matrix
mat3            nmat; // to be used instead of normal matrix
mat4            wave_rot_mat;
//-----------------------------------------------------------------
void wave_func(void); 
void build_norm_mat(void); 

//-----------------------------------------------------------------
void main(void)
{
    tp = vec3(0.0,0.0,0.0);
    tn = vec3(0.0,0.0,0.0);
    mvmat = gl_ModelViewMatrix;
    nmat = gl_NormalMatrix; 
    
    // apply wave displacement
    wave_func();
    
    // need to rebuild modelview matrix!!!
    mat4 tempmat; 

    // translation & scale
    tempmat = identity;
    tempmat[3][0] = 0.0;
    tempmat[3][1] = tp.y;
    tempmat[3][2] = testPosition.z;;
    mvmat *= tempmat;
  
    // scale the object
    tempmat = identity;
    tempmat[0][0] = ObjScale;
    tempmat[1][1] = ObjScale;
    tempmat[2][2] = ObjScale;
    mvmat *= tempmat;
 
    // rotate based on normal at wave
    mvmat *= wave_rot_mat;

    // rotate about local y axis
    tempmat = identity;
    tempmat[0][0] = cos(Spin);
    tempmat[2][0] = sin(Spin);
    tempmat[0][2] = -sin(Spin);
    tempmat[2][2] = cos(Spin);
    mvmat *= tempmat;
    
    // rebuild normal matrix for proper lighting
    build_norm_mat();
 
    // do brick shader stuff from book using our new mvmat and nmat
    vec4 ecPosition = mvmat * gl_Vertex;
    vec3 tnorm      = normalize( nmat * gl_Normal);
    vec3 lightVec   = normalize(vec3 (gl_LightSource[1].position) - 
                            ecPosition.xyz);
    vec3 reflectVec = reflect(-lightVec,tnorm);
    vec3 viewVec    = normalize(-ecPosition.xyz);
    float diffuse   = max(dot(lightVec,tnorm), 0.0);
    float spec      = 0.0;

    if( diffuse > 0.0 )
    {
        spec = max(dot(reflectVec,viewVec), 0.0);
        spec = pow(spec, 16.0);
    }

    LightIntensity  = DiffuseContribution * diffuse +
                        SpecularContribution * spec;

    // need to hand off model coordinate and normal to frag shader 
    MCposition      = gl_Vertex.xyz;
    normal          = normalize(gl_Normal);

    // ecPosition = mvmat * gl_Vertex 
    gl_Position     = gl_ProjectionMatrix * ecPosition;
}


//-----------------------------------------------------------------
float determinant(mat3 m)
{
    float t = 0.0;
    t = m[0][0] * ((m[1][1] * m[2][2]) - (m[2][1] * m[1][2]));
    t+= m[0][1] * ((m[1][2] * m[2][0]) - (m[2][2] * m[1][0]));
    t+= m[0][2] * ((m[1][0] * m[2][1]) - (m[2][0] * m[1][1]));

    return t;
}

float determinant(mat2 m)
{
    return ((m[0][0] * m[1][1]) - (m[1][0] * m[0][1]));
}

//-----------------------------------------------------------------
// normal matrix is transpose of inverse of upper left 3x3 of modelview matrix
void build_norm_mat(void) 
{
    mat3  tnmat;
    float d;
    for( int i = 0; i < 3; i++)
        for( int k = 0; k < 3; k++)
            nmat[i][k] = mvmat[i][k];
    
    d = determinant(nmat);
    // take determinant and transpose at same time
    tnmat[0][0] = 
        determinant(mat2(nmat[1][1], nmat[1][2], nmat[2][1], nmat[2][2]));
    tnmat[1][0] = 
        determinant(mat2(nmat[0][2], nmat[0][1], nmat[2][2], nmat[2][1]));
    tnmat[2][0] = 
        determinant(mat2(nmat[0][1], nmat[0][2], nmat[1][1], nmat[1][2]));
    tnmat[0][1] = 
        determinant(mat2(nmat[1][2], nmat[1][0], nmat[2][2], nmat[2][1]));
    tnmat[1][1] = 
        determinant(mat2(nmat[0][0], nmat[0][2], nmat[2][1], nmat[2][2]));
    tnmat[2][1] = 
        determinant(mat2(nmat[1][2], nmat[0][0], nmat[1][2], nmat[1][0]));
    tnmat[0][2] = 
        determinant(mat2(nmat[1][0], nmat[1][1], nmat[2][0], nmat[2][1]));
    tnmat[1][2] = 
        determinant(mat2(nmat[0][1], nmat[0][0], nmat[2][1], nmat[2][0]));
    tnmat[2][2] = 
        determinant(mat2(nmat[0][0], nmat[0][1], nmat[1][0], nmat[1][1]));

    nmat = tnmat / d;
}

//-----------------------------------------------------------------
void wave_func(void) 
{
    vec3 v1, v2, pos1, pos2;
    float h1, h2;
    
    // determine current height displacement 
    h1 = sin( (testPosition.z / zdiv) + Time) * magz; 
    h2 = sin( (testPosition.x / xdiv) + Time) * magx +
         sin( (testPosition.x / 5.0 ) + (Time - 4.0)); 
    tp = vec3( 0.0, h1 + h2, 0.0); // only take y component
    
    // get our two slopes in xy and zy
    // d/dx( sin(u) ) = cos(u)d/dx(u)
    // v1 with respect to z
    // v2 with respect to x
    v1 = vec3(  0.0, 
                cos((testPosition.z / zdiv) + Time) * magz / zdiv, 
                1.0);    
    v2 = vec3(  1.0, 
                cos((testPosition.x / xdiv) + Time) * magx / xdiv +
                cos( (testPosition.x / 5.0 ) + (Time - 4)) / 5.0, 
                0.0);   
    v1 = normalize(v1); 
    v2 = normalize(v2);

    // find our normal, should be in positive y direction
    tn = normalize(cross(v1,v2));
    
    // should have our basis vectors!! v1, v2, tn... now build a matrix
    wave_rot_mat = identity;
    wave_rot_mat[0] = vec4( v2, 0.0);
    wave_rot_mat[1] = vec4( tn, 0.0);
    wave_rot_mat[2] = vec4( v1, 0.0);
}
